/**
 * Copyright (c) 2010 jolira. All rights reserved. This program and the accompanying materials are made available under
 * the terms of the GNU Public License 2.0 which is available at http://www.gnu.org/licenses/old-licenses/gpl-2.0.html
 */
package com.google.code.joliratools.logdb.parser;

import static com.google.code.joliratools.logdb.parser.Record.Type.JUNK;
import static com.google.code.joliratools.logdb.parser.Record.Type.RECOGNIZED;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import com.google.code.joliratools.logdb.parser.Record.Type;

/**
 * Parse delimited records.
 * 
 * @author jfk
 * @date Oct 20, 2010 7:15:19 PM
 * @since 1.0
 */
public class DelimitedParser implements Parser {
    /**
     * The newline entry read by the newline.
     */
    final static String NEWLINE = System.getProperty("line.separator");

    private final BufferedReader reader;
    private int lineNo = 0;
    private final String fieldDelimiter;
    private StringBuilder buf = null;
    private int lastRecordStart = 0;
    private final String timeStampFormat;

    /**
     * the position of the timestamp in the list of fields
     */
    protected final int timeStampPos;

    /**
     * the position of the class name in the list of fields
     */
    protected final int classPos;
    /**
     * the position of the method name in the list of fields
     */
    protected final int methodPos;
    /**
     * the position of the level in the list of fields
     */
    protected final int levelPos;

    /**
     * Create a new parser.
     * 
     * @param in
     * @param fieldDelimiter
     * @param timeStampFormat
     * @param timeStampPos
     * @param classPos
     * @param methodPos
     * @param levelPos
     */
    protected DelimitedParser(final InputStream in, final String fieldDelimiter, final String timeStampFormat,
            final int timeStampPos, final int classPos, final int methodPos, final int levelPos) {
        this.fieldDelimiter = fieldDelimiter;
        this.timeStampFormat = timeStampFormat;
        this.timeStampPos = timeStampPos;
        this.classPos = classPos;
        this.methodPos = methodPos;
        this.levelPos = levelPos;

        reader = new BufferedReader(new InputStreamReader(in));
    }

    private void append(final String line) {
        if (buf == null) {
            buf = new StringBuilder();
        } else {
            buf.append(NEWLINE);
        }

        buf.append(line);
    }

    /**
     * Create a new record.
     * 
     * @param type
     *            the type recognized
     * @param lineNumber
     *            the current line number
     * @param line
     *            the line
     * @param fields
     *            the fields
     * @param timestamp
     *            the timestamp
     * @param clazz
     *            the class name
     * @param method
     *            the method name
     * @param level
     *            the level
     * @return the newly created record.
     */
    protected DelimitedRecord createRecord(final Type type, final int lineNumber, final String line,
            final List<String> fields, final long timestamp, final String clazz, final String method, final String level) {
        return new DelimitedRecord(type, lineNumber, line, fields, timestamp, clazz, method, level);
    }

    /**
     * get one particular field from the list of fields.
     * 
     * @param fields
     *            the fields found so far
     * @param pos
     *            the position of the field
     * @return return the parsed field or {@literal null}, if there is no such field
     * @throws IOException
     * @throws ParseException
     */
    protected String getField(final List<String> fields, final int pos) throws IOException, ParseException {
        final int size = fields.size();

        if (pos < 0 || pos >= size) {
            return null;
        }

        return fields.get(pos);
    }

    @Override
    public synchronized Record getNext() throws ParseException, IOException {
        for (;;) {
            final String line = readLine();

            if (line == null) {
                return parseBuffered();
            }

            final long timestamp = parseTimestamp(line);

            if (timestamp >= 0) {
                final Record _new = parseBuffered();

                lastRecordStart = lineNo;

                append(line);

                if (_new != null) {
                    return _new;
                }
            } else {
                append(line);
            }
        }
    }

    private Record parseBuffered() throws IOException, ParseException {
        if (buf == null) {
            return null;
        }

        final String line = buf.toString();

        buf = null;

        final List<String> fields = parseFields(line);
        final long timestamp = parseTimeStamp(fields);

        if (timestamp < 0) {
            return createRecord(JUNK, lastRecordStart, line, fields, timestamp, null, null, null);
        }

        final String clazz = getField(fields, classPos);
        final String method = getField(fields, methodPos);
        final String level = getField(fields, levelPos);

        return createRecord(RECOGNIZED, lastRecordStart, line, fields, timestamp, clazz, method, level);
    }

    private List<String> parseFields(final String line) {
        final List<String> vars = new ArrayList<String>();
        int start = 0;

        for (;;) {
            final int end = line.indexOf(fieldDelimiter, start);

            if (end == -1) {
                break;
            }

            final String field = start == end ? null : line.substring(start, end);

            vars.add(field);

            start = end + 1;
        }

        final String field = line.substring(start);

        vars.add(field);

        return vars;
    }

    private long parseTimestamp(final String line) throws IOException, ParseException {
        final List<String> fields = parseFields(line);

        return parseTimeStamp(fields);
    }

    /**
     * parse the timestamps.
     * 
     * @param fields
     *            the files
     * @return the timestamp
     * @throws IOException
     * @throws ParseException
     */
    protected long parseTimeStamp(final List<String> fields) throws IOException, ParseException {
        final String timeStamp = getField(fields, timeStampPos);

        if (timeStamp == null) {
            return -1;
        }

        final Date parsed = parseTimeStamp(timeStamp);

        return parsed != null ? parsed.getTime() : -1;
    }

    /**
     * Pare the timestamp.
     * 
     * @param timeStamp
     *            the time &amp; date
     * @return the date
     */
    protected Date parseTimeStamp(final String timeStamp) {
        final DateFormat format = new SimpleDateFormat(timeStampFormat);
        try {
            return format.parse(timeStamp);
        } catch (final java.text.ParseException e) {
            return null;
        }
    }

    private String readLine() throws IOException {
        final String line = reader.readLine();

        if (line != null) {
            lineNo++;
        }

        return line;
    }
}
